# Program to show more complex stack-based calling conventions

# Read in n
# Compute fib(n) very inefficiently by
#   fib(0) = 0;
#   fib(1) = 1;
#   fib(n) = fib(n-2) + fib(n-1) for n >=2
#
# n:   0 1 2 3 4 5 6  7
# Fib: 0 1 1 2 3 5 8 13

# loc 0 is always PC by hardware
# loc 1 is stack pointer sp by convention
# loc 2 is address of bottom of stack, 
#   grows upward to increasing addrs

# Idiom to initialize PC with START addr
_:START;

sp: stack;
addr_stack: stack;

# Constants, convention is _ddddd or _nddddd.  
# e.g. _12 for 12, _n2 for -2

_0: 0;
_1: 1;
_2: 2;
_3: 3;
_n1: -1;
_n2: -2;
_n3: -3;
_n8: -8;

# Debugging labels
debug_1:    1111;
debug_2:    2222;
debug_3:    3333;
debug_4:    4444;
debug_5:    5555;
debug_6:    6666;
debug_7:    7777;
debug_8:    8888;
debug_9:    9999;

# We compute m = fib(n)
n: 0; 
m: 0;

# count the number of calls
n_calls: 0;

# Temp variables, be careful not to use across a function call
# just like a register
tmp_r: 0;
tmp_n: 0;
result: 0;

# Always reset the stack pointer when you start
START: SUB, addr_stack, _0, sp;

# A stack frame on call looks like this:
# [[sp]-2] - function param n
# [[sp]-1] - return result fib(n)
# [[sp]] - return address

# During execution after entry and allocating local variables
# frame, the stack looks like this::
# [[sp]-4] - function param n
# [[sp]-3] - return result
# [[sp]-2] - return address
# [[sp]-1] - value of n on initial entry
# [[sp]]   - value of fib(n-2)

# Read n
SUB, IN, _0, n;

# Debug stack pointer before
SUB, sp, _0, OUT;

# Call fib(n), 
# Allocate space on the stack for param n, result m, return addr 
SUB, sp, _n3, sp;

# Put n into the param position at [sp]-2
# i.e. [ n ] -> [ [sp]-2 ]
SUB, sp, _2, @+7;
SUB, n, _0, 0;

# Put 0 into the result position at [sp]-1 to 0
# i.e. 0 -> [ [sp]-1 ]
SUB, sp, _1, @+7;
SUB, _0, _0, 0;

# Put return address (next instruction after the call) into 
# the return position at [sp].  
SUB, sp, _0, @+7;
SUB, PC, _n8, 0;

# Call the function
BGE, _0, _0, fib;
# Return here

# Fetch result
SUB, sp, _1, @+5;
SUB, 0, _0, m;

# Pop stack frame: param, result, and return address 
SUB, sp, _3, sp;

# Debug
SUB, debug_9, _0, OUT;

# Print n
SUB, n, _0, OUT;

# Print the result
SUB, m, _0, OUT;

# Print the the number of recursive calls, negated
SUB, _0, n_calls, OUT;

# Debug stack pointer after
SUB, sp, _0, OUT;

# Halt by setting PC to -PC
SUB, _0, PC, PC;
# Halt by setting PC to 0
SUB, PC, PC, PC;

#
# Body of function fib
#
fib:

# Count calls
SUB, n_calls, _n1, n_calls;

# Fetch the param into tmp_n, a global temp not safe over
# function calls
SUB, sp, _2, @+5;
SUB, 0, _0, tmp_n;

# Debug
SUB, debug_2, _0, OUT;
SUB, tmp_n, _0, OUT;
SUB, sp, _0, OUT;

# Only recurse if n > 2
BGE, _0, tmp_n, base_0;
BGE, _1, tmp_n, base_1;

# We need to store n and fib(n-2) across function calls.  So we 
# need to allocate two local variables on the stack to hold
# n 
# fib(n-2)

SUB, sp, _n2, sp;
SUB, sp, _0, @+7;
SUB, _0, _0, 0;
SUB, sp, _1, @+7;
SUB, tmp_n, _0, 0;

# Recursive call, put new param on stack

# Allocate space on the stack for param n, result m, return addr 
SUB, sp, _n3, sp;

# Put n-2 into the param position.
SUB, sp, _2, @+7;
SUB, tmp_n, _2, 0;

# Clear result location
SUB, sp, _1, @+7;
SUB, _0, _0, 0;

# Return address
SUB, sp, _0, @+7;
SUB, PC, _n8, 0;

# Call fib.
BGE, _0, _0, fib;
# return here 

# Fetch result fib(n-1)
SUB, sp, _1, @+5;
SUB, 0, _0, tmp_r;

# Pop stack frame
SUB, sp, _3, sp;

# Save fib(n-2) in local stack variable
SUB, sp, _0, @+7;
SUB, tmp_r, _0, 0;

# Get n-1
SUB, sp, _1, @+5;
SUB, 0, _1, tmp_n

# Debug
SUB, debug_3, _0, OUT;
SUB, tmp_r, _0, OUT;
SUB, tmp_n, _0, OUT;

# Call fib(n-1)

# Allocate space on the stack for param n, result m, return addr 
SUB, sp, _n3, sp;

# Put n-1 into the param position.
SUB, sp, _2, @+7;
SUB, tmp_n, _0, 0;

# Clear result location
SUB, sp, _1, @+7;
SUB, _0, _0, 0;

# Return address
SUB, sp, _0, @+7;
SUB, PC, _n8, 0;

# Call fib.
BGE, _0, _0, fib;
# return here 

# Fetch result of fib(n-1)
SUB, sp, _1, @+5;
SUB, 0, _0, result;

# Pop stack frame
SUB, sp, _3, sp;


# Fetch saved fib(n-2), negated
SUB, sp, _0, @+6;
SUB, _0, 0, tmp_r;

# Debug
SUB, debug_4, _0, OUT;
SUB, tmp_r, _0, OUT;
SUB, result, _0, OUT;

# fib(n-1) + fib(n-2)
SUB, result, tmp_r, result;

# Debug
SUB, debug_5, _0, OUT;
SUB, result, _0, OUT;

# Pop local variables
SUB, sp, _2, sp;

BGE, _0, _0, done;

base_1:

# Debug
SUB, debug_7, _1, OUT;

# n=1 Result is 1
SUB, _1, _0, result;
BGE, _0, _0, done;

base_0:
# Debug
SUB, debug_7, _0, OUT;

# n=0 Result is 0
SUB, _0, _0, result;
BGE, _0, _0, done;

# now return to the caller

done:
# Return, stack to be cleaned up by caller who is only 
# one that knows how many things were put on the stack.

# Debug
SUB, debug_8, _0, OUT;
SUB, result, _0, OUT;
SUB, sp, _0, OUT;

# Put returned result into stack, just below return address
SUB, sp, _1, @+7;
SUB, result, _0, 0;

# Get the return address from top of the stack
# and install it into the branch
SUB, sp, _0, @+5;
SUB, 0, _0, @+7;
BGE, _0, _0, 0;

# Stack starts here
stack: 0;
# Advance a bit so that the trace routine shows the stack cells
# with labels
@ = @ + 32
stack_end: 0;
