@Style{@Apalon{style/utility/source.s}}

@Center{
@Large{@B{The X-machine}} @L
H. James Hoover @L
Version 2.1 @L
2022-04-08
}

@P

When studying what it means to compute, the so-called theory of computation,
we often model the real hardware with abstractions called
"models of computation".  A model of computation captures the 
machine aspects that we want to explore, and simplifies away details that 
do not interest us at the time.

@P
Today we are going to look at a very simple model of computation, 
called the X-Machine.  I invented it back in my early days (1988) as a prof
to help students to understand the concepts of addresses and pointers in
CMPUT 330 (the earlier version of CMPUT 229, Dept. of Computing Science, University of Aleberta).  I found that programming the X-Machine
helped clarify the notions of memory, addresses, indexing, and linkage
conventions in preparation for the more complicated assembly language of the 
IBM 370 series machine.

@P
The X-machine is very simple.  It consists of memory, and a 
2-instruction CPU.

@P
@Large{@B{Memory}}

@P
Memory consists of an infinite sequence of cells, each cell contains an integer.
We will use decimal notation for writing our integers.  
The range of values in a cell is unlimited, in theory.  But in practice
we would have some idea of the largest magnitude that a cell would ever 
need.  

@SourcePage{doc-mem.src}

@P
We refer to the contents of a memory location using the notation [ ].
If the address of a location is A, then the contents of the location is
[A].  We allow addresses to be any integer value, including negative.
The notation "expr -> [ addr ]" means "set the contents of location
addr to the value of expr".

@P
For example
@Quote{
    [1] is the contents of memory location 1, and [1] = 1001 in our example above.
@L
@L
    [10] + 3 -> [11] means take the contents of location 10, add 3 to that value, and store it as the new contents of location [11]
@L
@L
    [1] + [[2]] -> [1] means increment the contents of location 1 by the
contents of the location whose address comes from the contents of location 2.
So in the above example [1] is 1001, [2] is 4, so [[2]] is 42. So the 
new value of location 1, i.e. [1], is 1043.
}

@Large{@B{Instruction Format}}
@P
Each instruction takes 4 memory cells.  The first cell is the opcode,
the next 3 cells are the operands.  There are two instructions, SUB
and BGE, with opcodes 1001 and 1002 respectively.

@Quote{
@Table{}
@Row{@| @B{Opcode} @| @B{Op1} @| @B{Op2} @| @B{Op3} @| @B{Action} }
@Row{@| SUB (1001) @| A @| B @| C @| [A] - [B] -> [C] }
@Row{@| BGE (1002) @| A @| B @| T @| if [A] >= [B] branch to T }
@Table{/}
}

So the instruction at address 1 above is: SUB, 4, 15, 42.  

@P
@Large{@B{Fetch Execute Cycle}}
@P

The advantage of this simple machine is that we can completely specify
its "operational semantics".  Memory location 0 is special, it holds the
current value of the program counter, which is the address in memory of
where the next instruction to be executed is located.  We refer to
it with the symbol PC, i.e. PC = 0.  We also define the symbols SUB = 1001
and BGE = 1002
@P
All memory cells with positive addresses contain 0 on boot, and are set 
to the values of the program before the machine is started.  This 
requires some magic - called the boot loader.
@P
Negative memory addresses are used for memory mapped IO.  Specific addresses
have special behaviour.

The symbols IN and CIN are read only, and for input.
The symbols OUT and COUT are write only, and for output.
In general IN and CIN are attached to stdin, and OUT and COUT are attached
to stdout.

Integer IO:
Reading from IN interprets the next characters from stdin as a signed integer.
Writing to OUT outputs the signed integer value to stdout as if it was printed
using the %d format.

Character IO:
Reading from CIN reads a single character from stdin and returns the 8bit ASCII
character code corresponding input character. Writing a value in [1, 255]
to COUT converts the integer into an 8bit ASCII character and sends it to stdout.
This operation has extended behaviour.  If the number is bigger than 255, then
it is broken into 8bit chunks, and the output as individual characters from
left to right.

@Quote{
while running @L
@Space{4}if [PC] == 0: halt @L
@L
@Space{4}@B{SUB, A, B, C i.e. [A] - [B] -> [C]} @L
@Space{4}if [ [PC] ] == SUB:  @L
@Space{8}[ [ [PC]+1 ] ] - [ [ [PC]+2 ] ] -> [ [ [PC]+3 ] ] @L
@Space{8}[PC] + 4 -> [PC] @L
@L
@Space{4}@B{BGE, A, B, T i.e. if [A] >= [B] branch to T} @L
@Space{4}if [PC] == BGE:  @L
@Space{8}if [ [ [PC]+1 ] ] >= [ [ [PC]+2 ] ] @L
@Space{12} then [ [PC]+3 ] -> [PC]  @L
@Space{12} else [PC]+4 -> [PC] @L
@L
@Space{4}otherwise: halt @L
}
Note the layers of fetching actual values to operate on! @L
[PC]+1 is the location of the first operand of the  instruction. @L
[ [PC] + 1 ] is the value of the operand, i.e. the address A. @L
and finally [ [ [PC] + 1 ] ] is the value [A] at address A.

@P
@Large{@B{Simple Sets of Instructions}}
@P

We now look at some simple sets of instructions.

@Points{

@+ Copy from location A to location B

@+ Increment location X by 1, i.e. [X] + 1 -> [X]

@+ Branch to location T

@+ Test if location X and location Y have the same values.  I.e.
if [X] = [Y] then do something, otherwise do something else.

@+ Compute the min of location X and Y and store in location Z, i.e. min( [X], [Y]) -> [Z]

@+ Relative branch forward 4 instructions

@+ Load the ith element of an array L
}

@L
@SourcePage{doc-simple-sets.xm}

@L
@SourcePage{doc-simple-sets.txt}

@P
@Large{@B{Example Programs}}
@P

@Large{Minimum of 2 integers}
@L
# @SourcePage{eg-min.xm}
@SourcePage{demo/min.txt}

@P 
@Large{Sum of 1+2+...+n}
@L
# @SourcePage{eg-sum.xm}
@SourcePage{demo/sum.txt}

@P 
@Large{Convert Non-negative to Binary}
@L
# @SourcePage{eg-binrep.xm}
@SourcePage{demo/binrep.txt}

@P 
@Large{Compute x div y and x mod y}
@L
# @SourcePage{eg-divide.xm}
@SourcePage{demo/divide.txt}

@P 
@Large{Recursive computation of 2 ** n}
@L
# @SourcePage{eg-recursion.xm}
@SourcePage{demo/recursion.txt}

@P 
@Large{Horrible recursive computation Fibonacci}
@L
# @SourcePage{eg-fib.xm}
@SourcePage{demo/fib.txt}

@L
@Large{@B{The Z-machine}}
@P

We could introduce a new 4-address instruction, SAB (subtract and branch), 
that combines the SUB and BGE instructions.

@Quote{
@Table{}
@Row{@| @B{Opcode} @| @B{Op1} @| @B{Op2} @| @B{Op3} @| @B{Op4} @| @B{Action} }
@Row{@| SAB (1003) @| A @| B @| C @| T @| 
    [A] - [B] -> [C] @L if [C] >= 0  branch to T }
@Table{/}
}

@P
SAB can implement SUB and BGE.  For SUB, just set the branch target T of
SAB to be the next instruction (T = @@+5).  
For BGE, throw away the result into a TMP location.

@Quote{
@Table{}
@Row{@| SUB, A, B, C @| => @| SAB, A, B, C, @@+5 }
@Row{@| BGE, A, B, T @| => @| SAB, A, B, TMP, T }
@Table{/}
}

But that means that we only need the SAB instruction.  We could build
a new machine, the Z-machine that has only one instruction.  So it does not
require an opcode - thus the Z for Zero opcode machine.  

So a program is just a string of addresses, every sequence of 4 address
being a legitimate instruction!  The SUB and BGE instructions
in the Z-machine are implemented as:

@Quote{
@Table{}
@Row{@| SUB, A, B, C @| => @|  A, B, C, @@+4 }
@Row{@| BGE, A, B, T @| => @|  A, B, TMP, T }
@Table{/}
}

This makes for an interesting code generation and optimization exercise.
