#!/usr/bin/perl
use strict;
use warnings;
use Scalar::Util;
use List::Util;

use Carp 'verbose';
$SIG{ __DIE__ } = \&Carp::confess;

use Data::Dumper;

my $debug = 0;

=head1 Simple X-machine tools: assembler, loader, and simulator

Version: 2.0, 2022-01-21

Usage:
    x-tool.pl [options]

where 

    -c file[.xm] - compile the X-machine assembly language program file.xm

    -o file[.xmout] - output an object file file.xmout suitable
        for loading with the boot loader.

    -l file[.xmout] - load the object file using the current boot loader

    -b file.xmout - load the boot loader code

    -io infile outfile - run the loaded X-machine code, attaching infile
        to the memory mapped input, and outfile to memory mapped output.
        If file is "-" use STDIN and STDOUT

    -run lim - start the program run for at most lim steps, 0 means
        run forever.

    -pc addr - override initial value in PC with addr.

    -d level - set debug level

    -t logfile - execution trace facility, sends trace to logfile

    -listing[.txt] file - put the program listing into file

note that only -c and -l are mutually exclusive

Assembly Language Syntax:

There are just two kinds of statements.

Definitions do not generate code, and are of the form:

    <id> = <expr> ;

They insert <id> into the symbol table with associated value of expression
<expr>.  

NOTE: If <id> is @, the current value of the location counter is
modified, which has the effect of re-origining where in memory the
instructions that follow are placed.

Instructions generate code in memory, and are of the form:

    [<id> : ] <expr_0>, <expr_1>, ..., <expr_n-1>;

An instruction has the following effects:

The "<id> :" part if present causes the value of <id> to be set to the 
current value of the location counter @.

The next n (n possibly 0) expressions are then placed in the next
n memory locations.

Any blank line, or line that starts with a # is considered to be a
comment and is ignored.

Semantics:

An identifier <id> can be inserted into the symbol table exactly once.
On insertion, it's value is the actual <expr>.

On mention, the value of an <id>is its associated <expr>.

As an instruction is generated, the address in memory where the 
it will be inserted has the <id> of "@" (as in "at" or "here").

=cut

sub throw 
{
    my ($message, @args) = @_;
    die sprintf ( $message, map { defined $_ ? $_ : "undef" } @args );
    }

package symtab;

sub new 
{
    my ($C, %A) = @_;

    my $I = bless { }, ref $C || $C;

    $I->{name} = { };
    $I->{reverse} = { };

    return $I;
    }

sub add 
{
    my ($I, $name, $value) = @_;
    if ( exists $I->{name}->{$name} ) {
        main::throw("Error: symbol %s already in symbol table with value '%s'.", 
            $name, $I->{name}->{$name} );
        }
    $I->{name}->{$name} = $value;
    }

sub update 
{
    my ($I, $name, $value) = @_;
    if ( ! exists $I->{name}->{$name} ) {
        $I->add($name, $value);
        }
    else {
        $I->{name}->{$name} = $value;
        }
    }

sub lookup 
{
    my ($I, $name) = @_;

    if ( ! exists $I->{name}->{$name} ) {
        main::throw("Error: symbol %s not in symbol table.", $name);
        }

    return $I->{name}->{$name};
    }

sub revidx
{
    my ($I) = @_;

    my $rev = $I->{reverse};
    foreach my $k ( keys %{$I->{name}} ) {
        my $val = $I->{name}->{$k};
        exists $rev->{$val} or
            $rev->{$val} = [ ];

        push @{$rev->{$val}}, $k;
        }
    }

sub revlookup 
{
    my ($I, $value) = @_;

    exists $I->{reverse}->{$value} or
        return [];

    return $I->{reverse}->{$value};
    }

sub dump 
{
    my ($I) = @_;
    return join("", 
        map { sprintf( "%4s: %s\n", 
             (defined $I->{name}->{$_} ? $I->{name}->{$_} : "undef"),
             $_ ) } 
            sort keys %{$I->{name}} );
    }

sub dump1
{
    my ($I) = @_;
    return join("", 
        map { "$_:" . (defined $I->{name}->{$_} ? $I->{name}->{$_} : "undef") . ":\n" } 
            sort keys %{$I->{Nmae}} );
    }

sub reduce_expr 
{
    my ($I, $expr) = @_;

    # Given a simple + - arithmetic expression involving numbers and
    # symbols, reduce by replacing symbols with their values, and then
    # if no symbols remaining, evaluate the expression to a number.

    $expr = $I->convert_strings( $expr );

    my @tokens = ( $expr =~ /([0-9]+|[()+\-*@]|[a-zA-Z0-9_]+)/g );

    # look up each token, and replace by value if possible
    @tokens = map { defined $I->{name}->{$_} ? $I->{name}->{$_} : $_ } @tokens;

    # convert back into an expression
    $expr = join(" ", @tokens);

    # if only numerics and operators left, evaluate it
    if ( $expr =~ /^[()+\-*\d\s]*$/ ) {
        my $r = eval($expr);
        $@ and main::throw("Error: could not evaluate expression |%s|", $expr);
        $expr = $r;
        }

    return $expr;
    }

sub convert_strings
{
    my ($I, $expr) = @_;

    # Replace short ascii string constants with their integer equivalents
    # so that they can be parsed as expressions.

    my $state = 'look_start_q';
    my $r = "";
    my $v = 0;

    my @chars = split //, $expr;

    CHAR: while ( @chars ) {
        my $c = shift @chars;

        if ( $state eq 'look_start_q' && $c ne '"' ) {
            $r .= $c;
            next CHAR;
            }

        if ( $state eq 'look_start_q' && $c eq '"' ) {
            $state = 'look_stop_q';
            $v = 0;
            next CHAR;
            }

        if ( $state eq 'look_stop_q' && $c ne '"' ) {
            # Handle escaped " and \
            if ( $c eq '\\' ) {
                $c = shift @chars;
                next CHAR unless defined $c;
                }
            $v = $v * 256 + ord($c);
            next CHAR;
            }

        if ( $state eq 'look_stop_q' && $c eq '"' ) {
            $r .= " $v ";
            $state = 'look_start_q';
            next CHAR;
            }
        }

    # If final state isn't look_start_q then we have unmatched quotes
    $state eq 'look_start_q'or
        main::throw "Mismatched '\"' in '%s'", $expr;
    

    return $r;
    }



package mcode;

use Data::Dumper;

sub new 
{
    my ($C, %A) = @_;

    # Create a new machine code object
    # symtab => user supplied symbol table, create fresh one if missing

    my $I = bless { }, ref $C || $C;

    $I->{pc} = 0;
    $I->{mem} = [ ];

    # Largest address encountered during assembly
    $I->{maxaddr} = 0;

    $I->{listing} = "";

    # if passed a symbol table use it, otherwise create a new one
    if ( $A{symtab} ) {
        $I->{symtab} = $A{symtab};
        }
    else {
        $I->{symtab} = symtab->new;
        }

    # start by installing specific symbols
    # current value of location counter
    $I->{symtab}->add('@', 0);

    return $I;
    }

sub insert 
{
    my ($I, $addr, $val) = @_;
    $I->{mem}->[$addr] = $val;
    $I->{maxaddr} = List::Util::max( $I->{maxaddr}, $addr );
    }

sub dump 
{
    my ($I) = @_;
    return Dumper($I);
    }

sub validate_id 
{
    my ($I, $id) = @_;

    # confirms that $id is a proper identifier, and removes any leading
    # or trailing whitespace.

    $id =~ s/^\s+//;
    $id =~ s/\s+$//;
    if ( $id =~ /\s/ ) {
        main::throw("Error: Id |%s| must not contain whitespace.", $id);
        }
    if ( $id ne '@' && $id !~ /^[a-zA-Z_][a-zA-Z_\d]*$/ ) {
        main::throw("Error: Id |%s| has invalid characters or format.", $id);
        }
    return $id; 
    }

sub assemble 
{
    my ($I, $target_loc, $statements) = @_;

    my $sym = $I->{symtab};

    # convert the input text into sequences of statements, possibly 
    # dropping comments

    my $drop_comments = 0;

    my @prog_stmts;

    LINE: foreach my $line ( @$statements ) {
        chomp($line); 

        if ( $drop_comments ) {
            next LINE if $line =~ /^\s*#/;
            next LINE if $line =~ /^\s*$/;
            }
        else {
            # keep comments and associate with the program statements
            if ( $line =~ /^\s*#/ ) {
                # remove leading whitespace
                $line =~ s/^\s+//;
                push @prog_stmts, $line;
                next LINE;
                }
            }

        # now process multi-line non-comment statments, and split into 
        # individual

        my @stmts = split /;/, $line;;

        # clean up non-comment statements
        foreach my $s ( @stmts ) {
            $s =~ s/^\s+//;
            $s =~ s/\s+$//;
            $s =~ s/\s+/ /g;
            }

        push @prog_stmts, @stmts;
        }

    my @prog_addr;
    my @prog_code;

    # next instruction location in memory
    my $loc = $target_loc;

    # now process each statment
    foreach my $s ( @prog_stmts ) {
        $debug and printf "Parsing |%s|\n", $s;

        # don't parse comments, but record them in the statements list
        if ( $s =~ /^\s*#/ ) {
            push @prog_code, $s;
            push @prog_addr, "";
            next;
            }

        # update @ symbol to next free memory slot where the
        # elements of the next statement will be installed.

        $sym->update('@', $loc);
        push @prog_addr, $sym->lookup('@');
        push @prog_code, [];

        if ( $s =~ /^\s*(.*)\s*=\s*(.*)$/ ) {
            # definition <id> = <expr>

            my $id = $I->validate_id($1);
            my $expr = $2;

            # immediately replace all instances of @ by the current value.
            $expr =~ s/@/$loc/gx;

            $debug and printf "id |%s| = |%s|\n", $id, $expr;

            # special case of updating the location counter
            if ( $id eq '@' ) {
                my $v = $sym->reduce_expr($expr);
    
                $debug and 
                    printf "setting loc from |%s| to |%s|\n", $expr, $v;

                if ( $v !~ /^\d+$/ ) {
                    main::throw
                        "Error: Could not reduce |%s| to address, got to |%s|",
                        $expr, $v;
                    }

                $sym->update('@', $v);
                $loc = $v;
                }
            else {
                $sym->add($id, $expr);
                }
            }
        elsif ( $s =~ /^.*[:]?$/ ) {
            my $id = "";
            my $rest = "";

            # fetch <id> if present
            if ( $s =~ /^\s*([^=]*)\s*:\s*(.*)$/ ) {
                # label present
                $id = $I->validate_id($1);
                $rest = $2;
                }
            else {
                $rest = $s;
                }

            # get <expr> list from rest of instruction
            my @exprs = split /,/, $rest;  

            $debug and
                printf "label |%s| : rest |%s|\n", $id, join("|", @exprs);
                
            if ( $id !~ /^\s*$/ ) {
                $sym->add($id, $loc);
                }

            foreach my $e ( @exprs ) {
                my $r = $sym->reduce_expr($e);

                $debug and printf "reducing |%s| to |%s|\n", $e, $r;

                $I->insert($loc, $r);
                # track the locations that are affected by this statement
                push @{$prog_code[-1]}, $loc;
                $loc++;
                }
            }
        else {
            throw("Error - Bad statement: |%s|\n", $s);
            }
        }

    # now go through and resolve any remaining expressions in mem
    # and set undef to 0;

    foreach my $m ( @{$I->{mem}} ) {
        defined $m or $m = 0;

        my $e = $m;

        $m = $sym->reduce_expr($m);
        # $m must be a number, or it was not resolved
        $m =~ /^-?\d+$/ or
            main::throw("Error: Could not resolve |%s| into a number, got as far as |%s|", $e, $m);
        }

    # generate the code listing
    $I->{listing} = "Code listing:\n";

    # comment indent amount
    my $indent = " " x 29;

    for (my $i=0; $i < @prog_stmts; $i++) {
        if ( $prog_code[$i] =~ /^\s*#/ ) {
            # Comments if present
            $I->{listing} .= 
                sprintf "%s%s\n", $indent, $prog_code[$i];
            next;
            }
        
        # $I->{listing} .= sprintf "%03d: %20s <- %s\n", 
        $I->{listing} .= sprintf "%03d: %20s    %s\n", 
            $prog_addr[$i], 
            join(" ", 
                map { sprintf("%4d", $I->{mem}->[$_]) } @{$prog_code[$i]}),
            $prog_stmts[$i];
        }
    }

sub dump_mem 
{
    # dump memory
    my ($I) = @_;

    my $mem = $I->{mem};

    my $result = "";
    $result .= sprintf "Memory contents\n";

    for (my $i=0; $i < @$mem; $i++) {
        my $v = $mem->[$i];
        defined $v or $v = "undef";
        $result .= sprintf "%2d: %s\n", $i, $v;
        }

    return $result;
    }

sub generate_xmout
{
    # Generate the xmout object stream to used by the boot loader
    #
    # returns a reference to a list

    my ($I) = @_;

    my $mem = $I->{mem};

    my $result = [ ];

    # xmout format:
    # start_addr
    # cell_count
    # cell contents for: start_addr ... start_addr + cell_count -1
    # 

    my $start = 0;
    my $count = $I->{maxaddr} - $start + 1;

    push @$result, $start;
    push @$result, $count;

    my $addr = $start;
    for (my $i=0; $i < $count; $i++) {
        my $v = $mem->[$addr];
        defined $v or $v = 0;
        push @$result, $v;
        $addr++;
        }


    return $result;
    }

package xmachine;

sub new 
{
    my ($C, %A) = @_;

    # Create a new xmachine 

    my $I = bless { }, ref $C || $C;

    $I->{code} = mcode->new( );

    # start by installing specific symbols representing the instruction set
    # of the X-machine, and special memory locations

    $I->{code}->{symtab}->add("SUB", 1001);
    $I->{code}->{symtab}->add("BGE", 1002);

    $I->{code}->{symtab}->add("PC", 0);
    $I->{code}->{symtab}->add("IN", -1);
    $I->{code}->{symtab}->add("OUT", -2);
    $I->{code}->{symtab}->add("CIN", -3);
    $I->{code}->{symtab}->add("COUT", -4);

    return $I;
    }

sub assemble 
{
    my ($I, $statements) = @_;

    $I->{code}->assemble(0, $statements);
    }

sub dump_mem
{
    my ($I, $statements) = @_;

    $I->{code}->dump_mem();
    }

sub delta_run 
{
    my ($I, %A) = @_;

    my @before_mem = @{$I->{code}->{mem}};
    $I->run(%A);
    my @after_mem = @{$I->{code}->{mem}};

    return (\@before_mem, \@after_mem);
    }

sub run 
{
    my ($I, %A) = @_;

    # Options:
    #   PC => use this start address instead of initial value in PC
    #   IN => input descriptor.  If undef, begin with STDIN.  If defined,
    #       is attached to a file, on EOF revert to STDIN.
    #   OUT => output descriptor, if defined attched to a file, otherwise
    #       use STDOUT.
    #   trace => 1 to enable memory access tracing
    #   trace_desc => FD for trace file 
    #   listing_desc => FD for listing file 
    #   max_steps => run for this many instruction steps, 0 means forever

    my $in_desc = $A{IN};
    my $in_state = defined $in_desc ? "FD" : "STDIN";

    $debug and printf STDERR "in_state '%s'\n", $in_state;

    my $out_desc = $A{OUT} //  \*STDOUT;
    # STDOUT->autoflush(1);

    my $start_addr = $A{PC} // $I->{mem}->[0];

    my $trace = $A{trace} // 0;
    my $trace_desc = $A{trace_desc} // \*STDERR;

    my $max_steps = $A{max_steps} // 0;

    my $code = $I->{code};
    my $symtab = $code->{symtab};

    # Lookup the predefined opcode and I/O symbol values

    my $opcode_SUB = $symtab->lookup("SUB");
    my $opcode_BGE = $symtab->lookup("BGE");

    my $io_IN = $symtab->lookup("IN");
    my $io_OUT = $symtab->lookup("OUT");
    my $io_CIN = $symtab->lookup("CIN");
    my $io_COUT = $symtab->lookup("COUT");

    $I->{num_steps} = 0;
    $I->{num_mem_reads} = 0;
    $I->{num_mem_writes} = 0;

    # run the program
    my $mem = $code->{mem};

    # $m->($addr) - get value [ $addr ]
    # $m->($addr, $v) - i.e. defined $v. set [ $addr ] = $v, return $v
    #
    # edge cases:
    #
    # undefined $addr => return 0
    #
    # $addr == IN => read in a value from the IN device, ignore
    #   set of value.  Returns 0 if EOF on IN.
    # $addr == OUT => write $v to the OUT device, return $v.
    # will return 0.  Do assignment if $v is defined.
    #
    # also record the last memory accesses for tracing purposes
    #
    # $tracelog->[idx]->[ [ idx, inst, pc, R|W, addr, value ] ]

    my $running = 1;

    $::SIG{INT} = sub {
        # On INT signal, stop running
        if ( $running ) {
            $running = 0;
            printf STDERR "INT caught, running is '%s'\n", $running;
            }
        else {
            main::throw "Aborting on second INT, %s steps", $I->{num_steps};
            }

        };

    my $tracelog = [ ];
    my $traceidx = 0;

    my $tr = sub {};

    if ( $trace ) {
        $tr = sub {
            my ($op, $addr, $v) = @_;

            my $pc = $mem->[0];

            push @$tracelog, [ $traceidx, $I->{num_steps}-1, $pc, $op, $addr, $v ];

            $traceidx++;
            };
        }

    # Pending input character buffer
    my $in_buf = "";

    my $m = sub {
        my ($addr, $v) = @_;

        $running or return 0;

        defined $addr or return 0;

        if ( defined $v ) {
            $I->{num_mem_writes}++;
            }
        else {
            $I->{num_mem_reads}++;
            }

        # Negative addresses are memory mapped io.  So every fetch from
        # IN returns a new value.  So need a tracing function that records
        # the memory information.

        # Special case of input being a tty, don't do STDIN fall back
        # from FD.  What should the behaviour be?  No fallback in general?

        if ( $addr < 0 ) {
            # Memory mapped IO handled here
            if ( $addr == $io_IN || $addr == $io_CIN ) {

                # When $in_buf is empty, do actual IO

                if ( $in_buf eq "" ) {
                    if ( $in_state eq "FD" ) {
                        # eof on a tty blocks waiting for an input
                        if ( (! -t $in_desc) && eof( $in_desc ) ) {
                            # IN file hit EOF, revert to STDIN
                            printf STDERR
                                "EOF encounterd on IN, swtiching to STDIN\n";

                            # main::throw "EOF encountered on IN";

                            close( $in_desc );

                            # Now divert to STDIN
                            $in_state = "STDIN";
                            }
                        else {
                            -t $in_desc and print $in_desc "?";
                            $in_buf = <$in_desc>;
                            defined $in_buf and chomp($in_buf);
                            }
                        }

                    if ( $in_state eq "STDIN" ) {
                        if ( eof( STDIN ) ) {
                            printf STDERR
                                "EOF encountered on STDIN, reads return 0\n";
                            $in_state = "EOF";
                            }
                        else {
                            # -t STDIN and print STDOUT "?\n";
                            $in_buf = <STDIN>;
                            chomp($in_buf);
                            }
                        }

                    if ( $in_state eq "EOF" ) {
                        $in_buf = "";
                        }

                    # At this point the input buffer has characters or
                    # is empty because of EOF.
                    }

                # What value is read depends on input mode
                if ( $in_buf eq "" ) {
                    $v = 0;
                    }
                elsif ( $addr == $io_IN ) {
                    $v = $in_buf;
                    $in_buf = "";
                    }
                elsif ( $addr == $io_CIN ) {
                    $v = ord( substr($in_buf, 0, 1) );
                    $in_buf = substr($in_buf, 1);
                    }

                $v = 0 unless defined $v;

                $tr->( "R", $addr, $v );

                Scalar::Util::looks_like_number( $v ) or
                    main::throw "Input stream has a non-number '%s'", $v;
                }

            elsif ( $addr == $io_OUT && defined $v ) {
                # Write integer value to OUT

                printf $out_desc "%s\n", $v;
                $tr->( "W", $addr, $v );
                }

            elsif ( $addr == $io_COUT && defined $v ) {
                # Write character values to OUT
                my $tv = $v;
                my $s = "";
                while ( $tv > 0 ) {
                    $s = chr($tv % 256) . $s;
                    $tv = int($tv / 256);
                    }

                # printf $out_desc "%d => %s\n", $v, $s;
                printf $out_desc "%s\n", $s;
                $tr->( "W", $addr, $v );
                }

            else {
                main::throw "Bad io register '%s'", $addr;
                }
            }
        else {
            # Normal memory handled here

            if ( defined $v ) {
                # Memory write

                $mem->[$addr] = $v;
                $tr->( "W", $addr, $v );

                }
            else {
                # Memory read

                $v = defined $mem->[$addr] ? $mem->[$addr] : 0;
                $tr->( "R", $addr, $v );
                }
            }

        return $v; 
        };


    # Initial valiue of PC is set by loader, but overridden by by
    # PC arg.

    defined $start_addr and
        $mem->[0] = $start_addr;

    printf STDERR "Starting at addr %d\n", $mem->[0];

    RUN: while ( $running ) {

        last RUN if $max_steps && $max_steps <= $I->{num_steps};

        $debug and printf STDERR "Step %s\n", $I->{num_steps};

        $I->{num_steps} ++;

        my $trace_enter = $traceidx;

        my $pc = $m->(0);

        if ( $pc <= 0 ) {
            # non-positive program counter means halt
            $running = 0;
            last RUN;
            }

        # Fetch opcode and the three addresses
        my $opcode = $m->($pc);
        defined $opcode or $opcode = 0;

        # fetch all
        
        if ( $opcode == $opcode_SUB ) {
            my $v;

            my $addr_A = $m->($pc+1);
            my $addr_B = $m->($pc+2);
            my $addr_C = $m->($pc+3);

            $v = $m->($addr_C, $m->($addr_A) - $m->($addr_B) );

            # if target address was PC, then don't increment it.
            if ( $addr_C != 0 ) {
                $m->(0, $m->(0)+4);
                }
            }

        elsif ( $opcode == $opcode_BGE ) {

            my $addr_A = $m->($pc+1);
            my $addr_B = $m->($pc+2);

            if ( $m->($addr_A) >= $m->($addr_B) ) {
                my $addr_C = $m->($pc+3);
                $m->(0, $addr_C);
                }
            else {
                $m->(0, $m->(0)+4);
                }

            }

        else {
            $running = 0;
            }

        my $trace_exit = $traceidx;

        if ( $trace ) {
            printf $trace_desc "%5s %5s %5s %s %5s %5s\n", 
                "IDX", "INST", "PC", "D", "ADDR", "VAL";
 
            for (my $i = $trace_enter; $i < $trace_exit; $i++ ) {
                printf $trace_desc "%5d %5d %5d %s %5d %5d\n", 
                    # map { $_ // "?" } @{$tracelog->[$i]};
                    map { $_ // "-42" } @{$tracelog->[$i]}[0..5];
                }

            printf $trace_desc "\n";
            }
        }

    printf STDERR "Halted after %d steps, %s max\n", 
        $I->{num_steps}, $max_steps ? $max_steps : "unlimited";
    printf STDERR "Num memory reads %d writes %s\n", 
        $I->{num_mem_reads}, $I->{num_mem_writes};
    }

sub dump_delta_mem 
{
    # dump before-after differences in memory
    my ($I, $before, $after) = @_;

    my $before_len = @$before;
    my $after_len = @$after;

    my $common_len = $before_len > $after_len ? $before_len : $after_len;

    my $result = "";
    $result .= sprintf "Before->After Memory Contents\n";

    # Compute the reverse index
    $I->{code}->{symtab}->revidx();

    for (my $i=0; $i < $common_len; $i++) {
        my $before_v = $before->[$i];

        my $after_v = $after->[$i];

        defined $before_v or $before_v = "undef";
        defined $after_v or $after_v = "undef";

        if ( $before_v ne $after_v ) {

            my $rev = $I->{code}->{symtab}->revlookup( $i );

            # Map addr to possible symbols with that addr
            my $revstr = "";

            if ( @$rev ) {
                $revstr = " (" . join(",", @$rev) . ")";
                }

            $result .= sprintf "%2d: %0s -> %0s%s\n", 
                $i, $before_v, $after_v, $revstr;
            }
        }

    return $result;
    }

sub generate_listing 
{
    my ($I) = @_;

    my $result = "";
    $result .= sprintf "\n%s\n", $I->{code}->{listing};

    # symbol table 
    $result .= sprintf "\nSymbol Table\n%s\n", $I->{code}->{symtab}->dump();

    return $result;
    }

package main;

# Options

my $trace = 0;

my $src_file;
my $xmout_file;
my $load_file;
my $boot_file;
my $trace_file;
my $listing_file;

my $in_file;
my $out_file;

my $do_run;
my $start_addr;
my $max_steps = 0;

if ( ! @ARGV ) {
    printf STDERR "See perldoc $0 for usage instructions\n";
    exit;
    }

OPTIONS: while ( @ARGV && $ARGV[0] =~ /^-/ ) {
    my $opt = shift @ARGV;
 
    if ( $opt eq "-t" ) {
        $trace = 1;
        $trace_file = shift @ARGV;

        defined $trace_file or 
            main::throw "No file supplied for -t option";
        
        $trace_file =~ /\.log$/ or
            $trace_file .= ".log";

        printf STDERR "Trace log file is '%s'\n", $trace_file;

        }

    elsif ( $opt eq "-c" ) {
        $src_file = shift @ARGV;

        defined $src_file or 
            main::throw "No file supplied for -c option";
        
        $src_file =~ /\.xm$/ or
            $src_file .= ".xm";

        printf STDERR "Source file is '%s'\n", $src_file;
        }

    elsif ( $opt eq "-listing" ) {
        $listing_file = shift @ARGV;

        defined $listing_file or 
            main::throw "No file supplied for -listing option";
        
        $listing_file =~ /\.txt$/ or
            $listing_file .= ".txt";

        printf STDERR "Listing file is '%s'\n", $listing_file;
        }

    elsif ( $opt eq "-o" ) {
        $xmout_file = shift @ARGV;

        defined $xmout_file or 
            main::throw "No file supplied for -o option";
        
        $xmout_file =~ /\.xmout$/ or
            $xmout_file .= ".xmout";

        printf STDERR "Object code will be in '%s'\n", $xmout_file;
        }

    elsif ( $opt eq "-l" ) {
        $load_file = shift @ARGV;

        defined $load_file or 
            main::throw "No file supplied for -l option";
        
        $load_file =~ /\.xmout$/ or
            $xmout_file .= ".xmout";

        printf STDERR "Object code from '%s'\n", $load_file;
        }

    elsif ( $opt eq "-b" ) {
        $boot_file = shift @ARGV;

        defined $boot_file or 
            main::throw "No file supplied for -b option";
        
        $boot_file =~ /\.xmout$/ or
            $xmout_file .= ".xmout";

        printf STDERR 
            "Boot loader being replaced by one from '%s'\n", $boot_file;
        }

    elsif ( $opt eq "-io" ) {
        $in_file = shift @ARGV;

        defined $in_file or 
            main::throw "No input file supplied for -io option";
        
        $out_file = shift @ARGV;

        defined $in_file or 
            main::throw "No output file supplied for -io option";
        
        printf STDERR 
            "Simulator running with input from '%s', output to '%s'\n",
            $in_file, $out_file;
        }

    elsif ( $opt eq "-pc" ) {
        $start_addr = shift @ARGV;

        defined $start_addr or 
            main::throw "No start addr supplied for -pc option";

        $start_addr > 0 or
            main::throw "start addr %s must be > 0 for -pc option", $start_addr;

        $do_run = 1;
        }

    elsif ( $opt eq "-run" ) {
        $max_steps = shift @ARGV;

        defined $max_steps or 
            main::throw "No step limit supplied for -run option";

        $do_run = 1;
        
        if ( $max_steps ) {
            printf STDERR "Simulator will run for at most %s steps\n", 
                $max_steps;
            }
        else {
            printf STDERR "Simulator will run forever\n";
            }
        }

    elsif ( $opt eq "-d" ) {
        $debug = shift @ARGV;

        defined $debug or 
            main::throw "No level supplied for -d option";

        printf STDERR "Debug level %s\n", $debug;
        }

    else {
        printf STDERR "Invalid option '%s'\n", $opt;
        exit;
        }
    }

printf STDERR "\n";
if ( defined $load_file && defined $src_file ) {
    printf STDERR "Cannot have both -l and -c options\n";
    exit 1;
    }

my $xmach = xmachine->new( );
   
# printf "%s\n", Dumper($xmach);

# fetch the input program source and assemble

my @statements;
if ( $src_file ) {
    open( my $SRC, "<", $src_file ) or
        main::throw("Cannot open source file '%s', %s", $src_file, $!);

    my $listing_desc;

    if ( defined $listing_file ) {
        open( $listing_desc, ">", $listing_file ) or
            main::throw("Cannot open listing file '%s', %s", $listing_file, $!);
        }
    else {
        $listing_desc = \*STDERR;
        }

    @statements = <$SRC>;

    close $SRC;

    $xmach->assemble( \@statements );

    printf $listing_desc "\n%s\n", $xmach->generate_listing();
    }

if ( $xmout_file ) {
    open( my $XMOUT, ">", $xmout_file ) or
        main::throw("Cannot open '%s', %s", $xmout_file, $!);

    my $result = $xmach->{code}->generate_xmout();

    foreach my $v ( @$result ) {
        printf $XMOUT "%s\n", $v;
        }

    close $XMOUT;
    }

my $in_desc;
my $out_desc;
my $trace_desc;

if ( $do_run ) {
    if ( defined $in_file && $in_file ne "-" ) {
        if ( $in_file eq "/dev/tty" ) {
            open( $in_desc, "+<", $in_file ) or
                main::throw("Cannot open input '%s', %s", $in_file, $!);
            }
        else {
            open( $in_desc, "<", $in_file ) or
                main::throw("Cannot open input '%s', %s", $in_file, $!);
            }
        }

    if ( defined $out_file && $out_file ne "-" ) {
        open( $out_desc, ">", $out_file ) or
            main::throw("Cannot open output '%s', %s", $out_file, $!);
        }

    if ( defined $trace_file ) {
        open( $trace_desc, ">", $trace_file ) or
            main::throw("Cannot open trace file '%s', %s", $trace_file, $!);
        }

    # Load object code if needed
    if ( $load_file ) {
        open( my $load_desc, "<", $load_file ) or
            main::throw("Cannot open object file '%s', %s", $load_file, $!);

        my $mem = $xmach->{code}->{mem};

        my $state = "LookAddr";
        my $addr;
        my $count;

        while ( my $l = <$load_desc> ) {
            chomp($l);
            if ( $state eq "LookAddr" ) {
                $addr = $l;
                $state = "LookCount";
                }
            elsif ( $state eq "LookCount" ) {
                $count = $l;
                if ( $count <= 0 ) {
                    # Count of zero means put addr into PC
                    $mem->[0] = $addr;
                    $state = "LookAddr";
                    $addr = 0;
                    $count = 0;
                    }
                else {
                    $state = "LookData";
                    }
                }
            elsif ( $state eq "LookData" ) {
                $mem->[$addr] = $l;
                $addr++;
                $count--;
                if ( $count <= 0 ) {
                    $state = "LookAddr";
                    $addr = 0;
                    $count = 0;
                    }
                }
            else {
                main::throw "Bad state '%s'", $state;
                }
            }

        }

    if ( 0 ) {
        $xmach->run();
        print $xmach->dump_mem();
        }
    else {
        # printf "\n%s\n", $xmach->dump_mem();

        # $in_desc->autoflush(1);

        printf "Running ...\n";
        my ($before, $after) = 
            $xmach->delta_run( 
                max_steps => $max_steps, 
                trace => $trace,
                trace_desc => $trace_desc,
                IN => $in_desc, 
                OUT => $out_desc, 
                PC => $start_addr,
                );

        printf "\n%s\n",  $xmach->dump_delta_mem($before, $after);

        printf "Program executed %s instructions\n", $xmach->{num_steps};
        printf "Num memory reads %d writes %s\n", 
            $xmach->{num_mem_reads}, $xmach->{num_mem_writes};
        }
    }
